Objects = {}
Objects_String = {}

function dist_to(x1, y1, x2, y2)
	x1 = x1 - x2
	y1 = y1 - y2
	return math.sqrt(x1 * x1 + y1 * y1)
end

function Objects_MakeString()
	Objects_String = {}
	local ninja_x, ninja_y
	for i,o in ipairs(Objects)	
		local what = o.what
		local x = o.x * 24 + o.xp
		local y = o.y * 24 + o.yp
		local s = ""
		
		if i == 1
			ninja_x = x
			ninja_y = y
		else
			local d = dist_to(x, y, ninja_x, ninja_y)
			while d < 8
				y = y - 1
				d = dist_to(x, y, ninja_x, ninja_y)
			end
		end
		
		if what == "ninja"
			s = "5^" .. x .. "," .. y
		elseif what == "gold"
			s = "0^" .. x .. "," .. y
		elseif what == "launchpad"
			s = "2^" .. x .. "," .. y .. "," .. o.xt .. "," .. o.yt
		elseif what == "gauss"
			s = "3^" .. x .. "," .. y
		elseif what == "door"
			local endings = {"0,0", "-1,0", "0,-1", "0,0"}
			s = "9^" .. x .. "," .. y .. "," .. o.orient .. ",0," .. o.dx .. "," .. o.dy .. ",1," .. endings[o.orient * 2 + o.where + 1]
		elseif what == "rocket"
			s = "10^" .. x .. "," .. y
		elseif what == "exit"
			s = "11^" .. x .. "," .. y .. "," .. (o.switch.x * 24 + o.switch.xp) .. "," .. (o.switch.y * 24 + o.switch.yp)
		elseif what == "mine"
			s = "12^" .. x .. "," .. y
		elseif what == "bounce"
			s = "1^" .. x .. "," .. y
		elseif what == "thuwmp"
			s = "8^" .. x .. "," .. y .. "," .. o.dir
		elseif what == "drone"
			s = "6^" .. table.concat( {x,y,o.algo,o.seeker,o.t,o.dir} , ",")
		end
		
		Objects_String[i] = s
	end
end

-- Ninja
function Objects_Phase1()
	local possible_positions = {}
	for id,region in ipairs(Terrain_Regions)
		if region.w >= 3 and region.h >= 3
			for x = region.x + 1, region.x + region.w - 2
				for y = region.y + 1, region.y + region.h - 2
					table.insert(possible_positions, {x=x,y=y})
				end
			end
		end
	end
	if #possible_positions == 0
		-- This is unlikely
		for x = 1, MAP_WIDTH
			for y = 1, MAP_HEIGHT
				if Terrain[x][y] == "0"
					table.insert(possible_positions, {x=x,y=y})
				end
			end
		end
	end
	if #possible_positions == 0
		-- This is virtually impossible
		for x = 1, MAP_WIDTH
			for y = 1, MAP_HEIGHT
				if Terrain[x][y] != "1"
					table.insert(possible_positions, {x=x,y=y})
				end
			end
		end
	end
	if #possible_positions == 0
		-- This is just retarded
		table.insert(possible_positions, {x=1,y=1})
	end
	local pos = possible_positions[math.random(1, #possible_positions)]
	
	-- Find distance down to something solid
	local d = 1
	while Terrain[pos.x][pos.y + d] == "0"
		d = d + 1
	end
	if d > 5
		pos.y = pos.y + d - 5
	end
	
	-- Mark regions as spawn zones
	d = 0
	while Terrain[pos.x][pos.y + d] != "1"
		local r = Terrain_Regions[Terrain_XY_To_Region[pos.x][pos.y + d]]
		r.spawn_zone = true
		r.no_mines = true
		d = d + 1
	end
	
	-- Add to object lists
	table.insert(Objects, {what="ninja",x=pos.x,y=pos.y,xp=12,yp=12})
	if FEEDBACK_SPEED == 1
		yield()
	end
end

-- Gold
function Objects_Phase2()
	local n = 1
	local corners = { {x=0,y=0},{x=1,y=0},{x=1,y=1},{x=0,y=1} }
	
	-- Put gold in closets
	for id,region in ipairs(Terrain_Regions)
		if region.openings == 1 and region.size == 1 and math.random(1,2) == 1
			for _,corner in corners
				if math.random(1,2) == 1
					table.insert(Objects, {
						what = "gold",
						x = region.x + (region.w - 1) * corner.x,
						y = region.y + (region.h - 1) * corner.y,
						xp = 18 - 12 * corner.x,
						yp = 18 - 12 * corner.y,
						region_id = id,
						closet_gold = true,
					})
					region.gold = true
					n = n + 1
					if (n % FEEDBACK_SPEED) == 0
						yield()
					end
				end
			end
		end
	end
	
	-- Put gold on walls
	local walls = {
		{s = [(r) | r.x], e = [(r) | r.x + r.w - 1], p = [(r,n) | n, r.y - 1  ], pp = [(n) | n * 6, 24]}, -- top
		{s = [(r) | r.x], e = [(r) | r.x + r.w - 1], p = [(r,n) | n, r.y + r.h], pp = [(n) | n * 6,  0]}, -- bottom
		{s = [(r) | r.y], e = [(r) | r.y + r.h - 1], p = [(r,n) | r.x - 1  , n], pp = [(n) | 24, n * 6]}, -- left
		{s = [(r) | r.y], e = [(r) | r.y + r.h - 1], p = [(r,n) | r.x + r.w, n], pp = [(n) |  0, n * 6]}, -- right
	}
	for id,region in ipairs(Terrain_Regions)
		if region.size >= 3
			for _,wall_def in walls
				local wall_good = true
				for n = wall_def.s(region),wall_def.e(region)
					local x,y = wall_def.p(region,n)
					if Terrain[x][y] != "1"
						wall_good = false
						break
					end
				end
				if wall_good
					wall_good = math.random(0, wall_def.e(region) - wall_def.s(region)) == 0
				end
				if wall_good
					local gap
					if (wall_def.e(region) - wall_def.s(region)) > 3
						gap = {2}
					else
						gap = {1,3}
					end
					for n = wall_def.s(region),wall_def.e(region)
						local x,y = wall_def.p(region,n)
						
						for _,g in gap
							local xp, yp = wall_def.pp(g)
							table.insert(Objects, {what = "gold", x=x, y=y, xp=xp, yp=yp, region_id = id, closet_gold = false})
							region.gold = true
							n = n + 1
							if (n % FEEDBACK_SPEED) == 0
								yield()
							end
						end
						
					end
				end
			end
		end
	end	
end

-- Rockets & gausses & drones
function Objects_Phase3()
	local n = 1
	local drones = {
		{t = 0, seeker = 0},
		{t = 0, seeker = 1},
		{t = 1, seeker = 0},
		{t = 2, seeker = 0},
	}
	local normal_drone = drones[1]
	local special_drone
	if math.random(1,5) <= 2
		special_drone = drones[2]
	elseif math.random(1,3) <= 2
		special_drone = drones[3]
	else
		special_drone = drones[4]
	end
	if math.random(1,7) == 1
		normal_drone = drones[2]
	end
	
	local function make_drone(region, details)
		table.insert(Objects, {
			what = "drone",
			x = math.random(region.x + 1, region.x + region.w - 2),
			y = math.random(region.y + 1, region.y + region.h - 2),
			xp = 12,
			yp = 12,
			algo = math.random(0, 5),
			seeker = details.seeker,
			t = details.t,
			dir = math.random(0,3),
		})
		n = n + 1
		if (n % FEEDBACK_SPEED) == 0
			yield()
		end
	end
	
	for id,region in ipairs(Terrain_Regions)
		if region.w >= 3 and region.h >= 3 and (not region.spawn_zone)
			if region.size > 42 and math.random(1,3) == 1
				table.insert(Objects, {what = "rocket", x=region.x, y=region.y, xp=region.w * 12, yp=region.h * 12})
				n = n + 1
				if (n % FEEDBACK_SPEED) == 0
					yield()
				end
			elseif math.random(1,4) == 1
				table.insert(Objects, {what = "gauss", x=region.x, y=region.y, xp=region.w * 12, yp=region.h * 12})
				n = n + 1
				if (n % FEEDBACK_SPEED) == 0
					yield()
				end
			elseif math.random(1,3) == 1
				make_drone(region, special_drone)
			elseif math.random(1,3) == 1
				make_drone(region, normal_drone)
			end
		end
	end
end

-- Bounce blocks
function Objects_Phase4()
	local n = 1
	for id,region in ipairs(Terrain_Regions)
		if region.h >= 4 and region.w >= 2
			local y = 0
			while (y + 4) <= region.h
				local sub_rgn = {x=region.x, y=region.y + y + 1, w=region.w, h=3}
				
				table.insert(Objects, 2, {
					what = "bounce",
					x = math.random(sub_rgn.x, sub_rgn.x + sub_rgn.w - 1),
					y = math.random(sub_rgn.y, sub_rgn.y + sub_rgn.h - 1),
					xp=12,
					yp=12,
				})
				n = n + 1
				if (n % FEEDBACK_SPEED) == 0
					yield()
				end
				
				y = y + 5
			end
		end
	end
end

-- Exits & doors
function Objects_Phase5()
	local max_dist = 0
	local n = 1
	Dist_From_Ninja = {}
	Edge_Cells = {}
	Empty_Cells = {}
	Walkable_Cells = {}
	Max_Dist_Cells = {}
	local Doors_To_Go_Through = {}
	local bad_for_switches = {["1"] = true, [";"] = true, [":"] = true, ["<"] = true, ["="] = true}
	
	for x = 1, MAP_WIDTH
		Dist_From_Ninja[x] = {}
		for y = 1, MAP_HEIGHT
			Dist_From_Ninja[x][y] = false
		end
	end
	
	local neighbours = { {x = -1,y = 0} , {x = 1, y = 0} , {x = 0, y = -1} , {x = 0, y = 1} }
	local list = { {x=Objects[1].x, y=Objects[1].y, d=0} }
	local list_itr = 1
	local lock_count = 0
	
	local function explore()
		while list[list_itr]
			local cell = list[list_itr]
			cell.lock_count = lock_count
			Dist_From_Ninja[cell.x][cell.y] = cell.d
			table.insert(Walkable_Cells, cell)
			
			if Terrain[cell.x][cell.y] == "0"
				table.insert(Empty_Cells, cell)
			end
			
			if cell.d > max_dist
				max_dist = cell.d
				Max_Dist_Cells = {}
			end
			if cell.d == max_dist
				table.insert(Max_Dist_Cells, cell)
			end
			
			local is_edge = true
			for _,neighbour in neighbours
				local nx = cell.x + neighbour.x
				local ny = cell.y + neighbour.y
				if Terrain[nx][ny] != "1" and Dist_From_Ninja[nx][ny] == false
					if Terrain_Regions[Terrain_XY_To_Region[nx][ny]].doorway
						table.insert(Doors_To_Go_Through, {x=nx, y=ny, d = cell.d + 1} )
					else
						table.insert(list, {x=nx, y=ny, d = cell.d + 1} )
					end
					is_edge = false
					Dist_From_Ninja[nx][ny] = cell.d + 1
				end
			end
			if is_edge
				if list_itr == 1
					lock_count = lock_count - 1
					cell.lock_count = lock_count
				end
				table.insert(Edge_Cells, cell)
			end
			list_itr = list_itr + 1
		end
	end
	explore()
	while Doors_To_Go_Through[1]
		local door = Doors_To_Go_Through[1]
		local orient = ((Terrain[door.x][door.y - 1] == "1") and 0) or 1
		local switch = Walkable_Cells[math.random(1, #Walkable_Cells)]
		
		while bad_for_switches[Terrain[switch.x][switch.y]]
			switch = Walkable_Cells[math.random(1, #Walkable_Cells)]
		end
		
		table.insert(Objects, {
			what = "door",
			x = switch.x,
			y = switch.y,
			xp = 12,
			yp = 12,
			dx = door.x,
			dy = door.y,
			orient = orient,
			where = math.random(0, 1),
		})
		
		Terrain_Regions[Terrain_XY_To_Region[door.x][door.y]].door_obj = Objects[#Objects]
		
		list = { door }
		list_itr = 1
		table.remove(Doors_To_Go_Through, 1)
		lock_count = lock_count + 1
		door.d = door.d + switch.d
		explore()
	end
	
	local Exit_Cells = {}
	for _,cell in Empty_Cells
		if Terrain[cell.x][cell.y + 1] == "1" and Terrain[cell.x - 1][cell.y + 1] == "1" and Terrain[cell.x + 1][cell.y + 1] == "1"
			table.insert(Exit_Cells, cell)
		end
	end
	if #Exit_Cells == 0
		Exit_Cells = Empty_Cells
	end
	
	local exit_cell = Exit_Cells[math.random(1, #Exit_Cells)]
	local switch_cell = Max_Dist_Cells[math.random(1, #Max_Dist_Cells)]
	
	while bad_for_switches[Terrain[switch_cell.x][switch_cell.y]]
		switch_cell = Max_Dist_Cells[math.random(1, #Max_Dist_Cells)]
	end
	
	Terrain_Regions[Terrain_XY_To_Region[exit_cell.x][exit_cell.y]].no_mines = true
	
	table.insert(Objects, 2, {
		what = "exit",
		x = exit_cell.x,
		y = exit_cell.y,
		xp = 12,
		yp = 12,
		switch = {
			x = switch_cell.x,
			y = switch_cell.y,
			xp = 12,
			yp = 12,
		}
	})
	
	n = n + 1
	if (n % FEEDBACK_SPEED) == 0
		yield()
	end
end

-- Mines & gold modifications
function Objects_Phase6()
	-- Find path from Ninja -> (doors) -> exit switch -> (doors) -> exit
	local objectives = {
		{x = Objects[1].x, y = Objects[1].y},
		{x = Objects[2].switch.x, y = Objects[2].switch.y},
		{x = Objects[2].x, y = Objects[2].y}
	}
	local cur_obj = 2
	local to_explore = {}
	local to_explore_itr = 1
	local explored = {}
	local opened_doors = {}
	local neighbours = { {x = -1,y = 0} , {x = 1, y = 0} , {x = 0, y = -1} , {x = 0, y = 1} }
	
	local function wipe_explored()
		explored = {}
		for x = 1, MAP_WIDTH
			explored[x] = {}
			for y = 1, MAP_HEIGHT
				explored[x][y] = false
			end
		end
	end
	
	while objectives[cur_obj]
		local target = objectives[cur_obj]
		to_explore = { objectives[cur_obj - 1] }
		to_explore_itr = 1
		wipe_explored()
		explored[objectives[cur_obj - 1].x][objectives[cur_obj - 1].y] = true
		
		while to_explore[to_explore_itr]
			local cell = to_explore[to_explore_itr]
			if cell.x == target.x and cell.y == target.y
				break
			end
			for _,neighbour in neighbours
				local nx = cell.x + neighbour.x
				local ny = cell.y + neighbour.y
				if Terrain[nx][ny] != "1" and (not explored[nx][ny])
					table.insert(to_explore, {x = nx, y = ny, from = cell})
					explored[nx][ny] = true
				end
			end
			to_explore_itr = to_explore_itr + 1
		end
		
		local first_door
		local cur_backtrack = to_explore[to_explore_itr]
		while cur_backtrack
			local region = Terrain_Regions[Terrain_XY_To_Region[cur_backtrack.x][cur_backtrack.y]]
			if region.doorway and (not opened_doors[region])
				first_door = region
			end
			cur_backtrack = cur_backtrack.from
		end
		if first_door
			table.insert(objectives, cur_obj, {x = first_door.door_obj.x, y = first_door.door_obj.y})
			opened_doors[first_door] = true
		else
			cur_backtrack = to_explore[to_explore_itr]
			while cur_backtrack
				Terrain_Regions[Terrain_XY_To_Region[cur_backtrack.x][cur_backtrack.y]].path_to_obj = cur_obj
				cur_backtrack = cur_backtrack.from
			end
			cur_obj = cur_obj + 1
		end
	end
	
	local n = 1
	local gold_to_redist = {}
	for i,o in ipairs(Objects)
		if o.what == "gold"
			local region = Terrain_Regions[o.region_id]
			if region.path_to_obj
				table.insert(gold_to_redist, o)
				if (region.w >= 2 and region.h >= 2 and region.size >= 5) and (not region.no_mines) and math.random(0, 1) == 1
					table.insert(Objects, {what = "mine", x=o.x, y=o.y, xp=o.xp, yp=o.yp})
					n = n + 1
					if (n % FEEDBACK_SPEED) == 0
						yield()
					end
				end
			end
		end
	end
	
	if #gold_to_redist > 0
		local redist_cells = {}
		for _,region in Terrain_Regions
			if not region.path_to_obj
				for x=region.x, region.x+region.w-1
					for y=region.y, region.y+region.h-1
						table.insert(redist_cells, {x=x, y=y})
					end
				end
			end
		end
		
		while #redist_cells > 0
			local for_this_cell = math.ceil(#gold_to_redist / #redist_cells)
			local cell = redist_cells[#redist_cells]
					
			for i=1,for_this_cell
				local gold = gold_to_redist[#gold_to_redist]
				gold.x = cell.x
				gold.y = cell.y
				table.remove(gold_to_redist)
				
				n = n + 1
				if (n % FEEDBACK_SPEED) == 0
					yield()
				end
			end
			table.remove(redist_cells)
		end
	end
end

-- Thuwmps
function Objects_Phase7()
	local update_n = 1
	
	local neighbours = { {x = 1,y = 0} , {x = 0, y = 1} , {x = -1, y = 0} , {x = 0, y = -1} }
	local surrounds = {}
	for i,n in neighbours
		if n.x == 0
			surrounds[i] = { {-1,-n.y},{1,-n.y} }
		else
			surrounds[i] = { {-n.x,-1},{-n.x,1} }
		end
	end
	for _,cell in Empty_Cells
		local solid_count = 0
		local non_solid_i = false
		
		for i,n in neighbours
			local t = Terrain[cell.x + n.x][cell.y + n.y]
			if t == "0"
				non_solid_i = i
			elseif t == "1"
				solid_count = solid_count + 1
			end
		end
		
		if solid_count == 3 and non_solid_i
			solid_count = 0
			for i,s in surrounds[non_solid_i]
				if s != "0"
					solid_count = solid_count + 1
				end
			end
			
			if solid_count == 2
				solid_count = 0
				local x,y = cell.x,cell.y
				while Terrain[x][y] == "0"
					if Terrain_Regions[Terrain_XY_To_Region[x][y]].spawn_zone
						solid_count = 0
						break
					end
					x = x + neighbours[non_solid_i].x
					y = y + neighbours[non_solid_i].y
					solid_count = solid_count + 1
				end
				
				if solid_count >= 4 and math.random(1,3) == 1
					table.insert(Objects, {what = "thuwmp", x=cell.x, y=cell.y, xp=12, yp=12, dir=non_solid_i-1})
					update_n = update_n + 1
					if (update_n % FEEDBACK_SPEED) == 0
						yield()
					end
				end
			end
		end
	end
end

-- Launch pads
function Objects_Phase8()
	local update_n = 1
	
	for _,cell in Empty_Cells
		if Terrain[cell.x][cell.y + 1] == "1"
			local empty_count = 0
			while Terrain[cell.x][cell.y - empty_count - 1] == "0"
				empty_count = empty_count + 1
			end
			if empty_count >= 7 and math.random(1,25) <= empty_count
				table.insert(Objects, {what = "launchpad", x=cell.x, y=cell.y, xp=12, yp=24, xt=0, yt=-1})
				update_n = update_n + 1
				if (update_n % FEEDBACK_SPEED) == 0
					yield()
				end
			end
		end
	end
end
