TerrainInfo = {}
TerrainInfo["0"] = {0,0,0,0,0,0,0,0}
TerrainInfo["1"] = {2,2,2,2,2,2,2,2}
TerrainInfo["2"] = {1,1,1,1,2,2,2,2}
TerrainInfo["3"] = {1,1,2,2,2,2,1,1}
TerrainInfo["4"] = {2,2,2,2,1,1,1,1}
TerrainInfo["5"] = {2,2,1,1,1,1,2,2}
TerrainInfo["6"] = {1,1,1,1,2,2,2,2}
TerrainInfo["7"] = {1,1,2,2,2,2,1,1}
TerrainInfo["8"] = {2,2,2,2,1,1,1,1}
TerrainInfo["9"] = {2,2,1,1,1,1,2,2}
TerrainInfo[":"] = {1,1,1,1,2,2,2,2}
TerrainInfo[";"] = {1,1,2,2,2,2,1,1}
TerrainInfo["<"] = {2,2,2,2,1,1,1,1}
TerrainInfo["="] = {2,2,1,1,1,1,2,2}
TerrainInfo[">"] = {1,1,0,1,2,2,2,0}
TerrainInfo["?"] = {1,1,0,2,2,2,1,0}
TerrainInfo["@"] = {2,2,2,0,1,1,0,1}
TerrainInfo["A"] = {2,2,1,0,1,1,0,2}
TerrainInfo["B"] = {1,1,1,2,2,2,2,2}
TerrainInfo["C"] = {1,1,2,2,2,2,2,1}
TerrainInfo["D"] = {2,2,2,2,1,1,1,2}
TerrainInfo["E"] = {2,2,2,1,1,1,2,2}
TerrainInfo["F"] = {1,0,1,1,0,2,2,2}
TerrainInfo["G"] = {0,1,2,2,2,0,1,1}
TerrainInfo["H"] = {0,2,2,2,1,0,1,1}
TerrainInfo["I"] = {2,0,1,1,0,1,2,2}
TerrainInfo["J"] = {2,1,1,1,2,2,2,2}
TerrainInfo["K"] = {1,2,2,2,2,2,1,1}
TerrainInfo["L"] = {2,2,2,2,2,1,1,1}
TerrainInfo["M"] = {2,2,1,1,1,2,2,2}
TerrainInfo["N"] = {2,2,2,0,1,1,0,2}
TerrainInfo["O"] = {2,0,1,1,0,2,2,2}
TerrainInfo["P"] = {1,1,0,2,2,2,2,0}
TerrainInfo["Q"] = {0,2,2,2,2,0,1,1}

TerrainSets = {"2345", "6789:;<=", ">?@ABCDEFGHIJKLMNOPQ", "2345>?@ABCDEFGHIJKLMNOPQ"}
TerrainCorners = {"5432o", "9876g", "A@?>g", "IHGFg", "=<;:o"} -- o opposing / g good

Terrain = setmetatable({}, {__tostring = function(t)
	local r = ""
	for x=1,MAP_WIDTH
		for y=1,MAP_HEIGHT
			r = r .. t[x][y]
		end
	end
	return r
end})
for x = 0, MAP_WIDTH + 1
	Terrain[x] = {}
	for y = 0, MAP_HEIGHT + 1
		if x == 0 or x == MAP_WIDTH + 1 or y == 0 or y == MAP_HEIGHT + 1
			Terrain[x][y] = "1"
		else
			Terrain[x][y] = "0"
		end
	end
end

-- Place random splats
function Terrain_Phase1()
	local upper = math.sqrt(MAP_WIDTH * MAP_HEIGHT):floor()
	local limit = math.random(upper / 2, upper)
	for i = 1, limit
		Terrain[math.random(1, MAP_WIDTH)][math.random(1, MAP_HEIGHT)] = "1"
		if (i % FEEDBACK_SPEED) == 0
			yield()
		end
	end
end

-- Grow random splats
function Terrain_Phase2()
	local solids = {}
	for x = 1, MAP_WIDTH
		for y = 1, MAP_HEIGHT
			if Terrain[x][y] == "1"
				table.insert(solids, {x = x, y = y, r = 0.72})
				table.insert(solids, {x = x, y = 0, r = 0.42})
				table.insert(solids, {x = x, y = MAP_HEIGHT + 1, r = 0.42})
				table.insert(solids, {x = 0, y = y, r = 0.42})
				table.insert(solids, {x = MAP_WIDTH + 1, y = y, r = 0.42})
			end
		end
	end
	local i = 1
	local n = 1
	while solids[i]
		local neighbours = { {x = -1,y = 0} , {x = 1, y = 0} , {x = 0, y = -1} , {x = 0, y = 1} }
		for _,neighbour in neighbours
			local x = solids[i].x + neighbour.x
			local y = solids[i].y + neighbour.y
			local r = solids[i].r
			
			if Terrain[x] and Terrain[x][y] == "0" and math.random() < r
				Terrain[x][y] = "1"
				table.insert(solids, {x = x, y = y, r = r - 0.09})
				n = n + 1
				if (n % FEEDBACK_SPEED) == 0
					yield()
				end
			end
		end
		i = i + 1
	end
end

-- Trim random splats
function Terrain_Phase3()
	local groups = {}
	for x = 0, MAP_WIDTH + 1
		groups[x] = {}
		for y = 0, MAP_HEIGHT + 1
			groups[x][y] = 0
		end
	end
	
	local n = 1
	local group_sizes = {}
	local next_group = 1
	local neighbours = { {x = -1,y = 0} , {x = 1, y = 0} , {x = 0, y = -1} , {x = 0, y = 1} }
	local neighbours_ext = { {x = -1,y = 0} , {x = 1, y = 0} , {x = 0, y = -1} , {x = 0, y = 1}
		, {x = -1, y = -1} , {x = -1, y = 1} , {x = 1, y= -1}, {x = 1, y = 1} }
	
	for x = 1, MAP_WIDTH
		for y = 1, MAP_HEIGHT
			if Terrain[x][y] == "0" and groups[x][y] == 0
				groups[x][y] = next_group
				group_sizes[next_group] = 1
				local explore = { {x=x,y=y} }
				local i = 1
				while explore[i]
					for _,neighbour in neighbours
						local x = explore[i].x + neighbour.x
						local y = explore[i].y + neighbour.y
						if Terrain[x] and Terrain[x][y] == "0" and groups[x][y] == 0
							groups[x][y] = next_group
							group_sizes[next_group] = group_sizes[next_group] + 1
							table.insert(explore, {x=x, y=y})
						end
					end
					i = i + 1
				end
				next_group = next_group + 1
			end
		end
	end
	
	for x = 1, MAP_WIDTH
		for y = 1, MAP_HEIGHT
			if Terrain[x][y] == "1"
				local neighbour_groups = {}
				neighbour_groups[0] = true
				local neighbour_group_count = 0
				
				for _,neighbour in neighbours_ext
					local nx = x + neighbour.x
					local ny = y + neighbour.y
					local ng = groups[nx][ny]
					if not neighbour_groups[ng]
						neighbour_groups[ng] = true
						neighbour_group_count = neighbour_group_count + 1
					end
				end
				
				if neighbour_group_count > 1
					Terrain[x][y] = "0"
					
					n = n + 1
					if (n % FEEDBACK_SPEED) == 0
						yield()
					end
				end
			end
		end
	end
	
	if n > 1
		return Terrain_Phase3()
	end
	
	local largest_group = 1
	local largest_size = group_sizes[1]
	for grp,size in group_sizes
		if size > largest_size
			largest_group = grp
			largest_size = size
		end
	end
	
	local n = 1
	for x = 1, MAP_WIDTH
		for y = 1, MAP_HEIGHT
			if Terrain[x][y] != "1" and groups[x][y] != largest_group
				Terrain[x][y] = "1"
				
				n = n + 1
				if (n % FEEDBACK_SPEED) == 0
					yield()
				end
			end
		end
	end
end

-- Identify areas
function Terrain_Phase4()
	local areas = {}
	Terrain_Regions = {}
	for x = 0, MAP_WIDTH + 1
		areas[x] = {}
		for y = 0, MAP_HEIGHT + 1
			areas[x][y] = 0
		end
	end
		
	for x = 1, MAP_WIDTH
		for y = 1, MAP_HEIGHT
			if Terrain[x][y] == "1"
				areas[x][y] = -1
			end
		end
	end
	
	local is_region_valid = function(rgn)
		for x = rgn.x, rgn.x+rgn.w-1
			for y = rgn.y, rgn.y+rgn.h-1
				if x < 1 or y < 1 or x > MAP_WIDTH or y > MAP_HEIGHT
					return false
				end
				if areas[x][y] != 0
					return false
				end
			end
		end
		return true
	end
	local next_rgn_id = 1
	while true
		local best_rgn = nil
		for x = 1, MAP_WIDTH
			for y = 1, MAP_HEIGHT
				if areas[x][y] == 0
					local rgn = {x=x, y=y, w=1, h=1}
					local grew = true
					while grew
						grew = false
						rgn.w = rgn.w + 1
						if is_region_valid(rgn)
							grew = true
						else
							rgn.w = rgn.w - 1
						end
						rgn.h = rgn.h + 1
						if is_region_valid(rgn)
							grew = true
						else
							rgn.h = rgn.h - 1
						end
					end
					rgn.score = (rgn.w * rgn.h) / (rgn.w + rgn.h + rgn.w + rgn.h)
					if (not best_rgn) or rgn.score > best_rgn.score
						best_rgn = rgn
					end
				end
			end
		end
		if not best_rgn
			break
		end
		for x = best_rgn.x, best_rgn.x+best_rgn.w-1
			for y = best_rgn.y, best_rgn.y+best_rgn.h-1
				areas[x][y] = next_rgn_id
			end
		end
		best_rgn.size = best_rgn.w * best_rgn.h
		best_rgn.doorway = false
		Terrain_Regions[next_rgn_id] = best_rgn
		
		local neighbours = {}
		for x = best_rgn.x, best_rgn.x+best_rgn.w
			table.insert(neighbours, {x = x, y = best_rgn.y - 1})
		end
		for y = best_rgn.y, best_rgn.y+best_rgn.h
			table.insert(neighbours, {x = best_rgn.x + best_rgn.w, y = y})
		end
		for x = best_rgn.x+best_rgn.w-1, best_rgn.x-1, -1
			table.insert(neighbours, {x = x, y = best_rgn.y + best_rgn.h})
		end
		for y = best_rgn.y+best_rgn.h-1, best_rgn.y-1, -1
			table.insert(neighbours, {x = best_rgn.x - 1, y = y})
		end
		neighbours[0] = neighbours[ #neighbours ]
		Terrain_Regions[next_rgn_id].neighbours = neighbours
		Terrain_Regions[next_rgn_id].openings = 0
		
		for i,n in ipairs(neighbours)
			if Terrain[n.x][n.y] != Terrain[ neighbours[i-1].x ][ neighbours[i-1].y ]
				Terrain_Regions[next_rgn_id].openings = Terrain_Regions[next_rgn_id].openings + 1
			end
		end
		Terrain_Regions[next_rgn_id].openings = Terrain_Regions[next_rgn_id].openings / 2
		Terrain_Regions[next_rgn_id].tunnel = ((best_rgn.w == 1) or (best_rgn.h == 1)) and (Terrain_Regions[next_rgn_id].openings > 1)
		if Terrain_Regions[next_rgn_id].tunnel and best_rgn.size == 1
			if (Terrain[best_rgn.x][best_rgn.y - 1] == "1" and
				Terrain[best_rgn.x + 1][best_rgn.y] != "1" and
				Terrain[best_rgn.x][best_rgn.y + 1] == "1" and
				Terrain[best_rgn.x - 1][best_rgn.y] != "1") or
				(Terrain[best_rgn.x][best_rgn.y - 1] != "1" and
				Terrain[best_rgn.x + 1][best_rgn.y] == "1" and
				Terrain[best_rgn.x][best_rgn.y + 1] != "1" and
				Terrain[best_rgn.x - 1][best_rgn.y] == "1") then
					best_rgn.doorway = true
			end
		end
		
		if log
			log:write("Region " ..  string.char( string.byte("A") + next_rgn_id - 1 ) .. ": " .. #neighbours .. "n " .. Terrain_Regions[next_rgn_id].openings .. "o \n")
		end
		
		next_rgn_id = next_rgn_id + 1
	end
	
	Terrain_XY_To_Region = areas
end

-- Apply decor to region corners
function Terrain_Phase5()
	TerrainSet = math.random(1, #TerrainSets)
	local n = 1
	while TerrainCorners[n]
		if TerrainSets[TerrainSet]:find(TerrainCorners[n]:sub(1,1), 1, true)
			n = n + 1
		else
			table.remove(TerrainCorners, n)
		end
	end
	
	local n = 1
	for id,region in ipairs(Terrain_Regions)
		if region.w >= 2 and region.h >= 2
			local hard_corners = {}
			local corners = { {x=0,y=0},{x=1,y=0},{x=1,y=1},{x=0,y=1} }
			for c,corner in corners
				local x = region.x + (region.w - 1) * corner.x
				local y = region.y + (region.h - 1) * corner.y
				
				if Terrain[x][y - 1 + corner.y * 2] == "1" and Terrain[x - 1 + corner.x * 2][y] == "1"
					table.insert(hard_corners, {x=x,y=y,c=c})
				end
			end
			
			if #hard_corners >= 1
				local corner_set = math.random(1, #TerrainCorners)
				
				for _,corner in hard_corners
					Terrain[corner.x][corner.y] = TerrainCorners[corner_set]:sub(corner.c, corner.c)
					
					n = n + 1
					if (n % FEEDBACK_SPEED) == 0
						yield()
					end
				end
			end
		elseif region.w == 1 or region.h == 1
			local hard_corners = {}
			local corners = { {x=0,y=0},{x=1,y=0},{x=1,y=1},{x=0,y=1} }
			for c,corner in corners
				local x = region.x + (region.w - 1) * corner.x
				local y = region.y + (region.h - 1) * corner.y

				if Terrain[x][y - 1 + corner.y * 2] == "1" and Terrain[x - 1 + corner.x * 2][y] == "1" and
				  Terrain[x][y + 1 - corner.y * 2] != "1" and Terrain[x + 1 - corner.x * 2][y] != "1"
					table.insert(hard_corners, {x=x,y=y,c=c})
				end
			end
			
			if #hard_corners >= 1
				local corner_set = math.random(1, #TerrainCorners)
				for n=1,4
					if TerrainCorners[corner_set]:sub(5,5) != "g"
						corner_set = math.random(1, #TerrainCorners)
					end
				end
				if TerrainCorners[corner_set]:sub(5,5) == "g"
					for _,corner in hard_corners
						Terrain[corner.x][corner.y] = TerrainCorners[corner_set]:sub(corner.c, corner.c)
						
						n = n + 1
						if (n % FEEDBACK_SPEED) == 0
							yield()
						end
					end
				end
			end
		end
	end
end
