/* 
	Corsix's N Level Generator
    Copyright (C) 2007 Peter Cawley

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef _TERRAIN_H_
#define _TERRAIN_H_

#include "wx/wxprec.h"
#ifdef __BORLANDC__
    #pragma hdrstop
#endif
#ifndef WX_PRECOMP
    #include "wx/app.h"
    #include "wx/log.h"
    #include "wx/frame.h"
    #include "wx/panel.h"
    #include "wx/timer.h"
    #include "wx/utils.h"
    #include "wx/menu.h"
    #include "wx/msgdlg.h"
    #include "wx/textdlg.h"
    #include "wx/filedlg.h"
    #include "wx/choicdlg.h"
    #include "wx/button.h"
    #include "wx/textctrl.h"
    #include "wx/listbox.h"
    #include "wx/sizer.h"
#endif
#include "wx/txtstrm.h"
#include "wx/numdlg.h"
#include "wx/ffile.h"
#include "wx/process.h"

#include "wx/bitmap.h"
#include "wx/dc.h"
#include <vector>

class Terrain
{
public:
	Terrain();
	~Terrain();

	bool generateTileImages();
	bool loadObjectImages();

	bool renderOnto(wxDC& oDest);

	void setOverlay(const char* sOverlayData);
	void load(const char* sMapData);

	static const int k_iWidth    = 33;
	static const int k_iHeight   = 25;
	static const int k_iCellSize = 24;

protected:
	int* m_pCellTiles;
	wxString* m_pCellTilesOverlay;

	wxBitmap m_bmpTiles;
	wxBitmap m_bmpObjects;
	wxBitmap m_bmpObjects2;

	struct _object_t
	{
		virtual int getSheetIndex();
		virtual bool render(wxDC& oDest, wxDC* pObjectSheet) = 0;
		virtual void load(const char* sData, size_t iLen) = 0;
	};

	struct _ninja_t : _object_t
	{
		virtual bool render(wxDC& oDest, wxDC* pObjectSheet);
		virtual void load(const char* sData, size_t iLen);

		int x,y;
	};

	struct _gold_t : _object_t
	{
		virtual bool render(wxDC& oDest, wxDC* pObjectSheet);
		virtual void load(const char* sData, size_t iLen);

		int x,y;
	};

	struct _mine_t : _object_t
	{
		virtual bool render(wxDC& oDest, wxDC* pObjectSheet);
		virtual void load(const char* sData, size_t iLen);

		int x,y;
	};

	struct _rocket_t : _object_t
	{
		virtual bool render(wxDC& oDest, wxDC* pObjectSheet);
		virtual void load(const char* sData, size_t iLen);

		int x,y;
	};

	struct _gauss_t : _object_t
	{
		virtual bool render(wxDC& oDest, wxDC* pObjectSheet);
		virtual void load(const char* sData, size_t iLen);

		int x,y;
	};

	struct _bounce_t : _object_t
	{
		virtual bool render(wxDC& oDest, wxDC* pObjectSheet);
		virtual void load(const char* sData, size_t iLen);

		int x,y;
	};

	struct _launchpad_t : _object_t
	{
		virtual bool render(wxDC& oDest, wxDC* pObjectSheet);
		virtual void load(const char* sData, size_t iLen);

		int x,y;
		double xt, yt;
	};

	struct _thuwmp_t : _object_t
	{
		virtual bool render(wxDC& oDest, wxDC* pObjectSheet);
		virtual void load(const char* sData, size_t iLen);

		int x,y,d;
	};

	struct _exit_t : _object_t
	{
		virtual bool render(wxDC& oDest, wxDC* pObjectSheet);
		virtual void load(const char* sData, size_t iLen);

		int x,y, sx, sy;
	};

	struct _door_t : _object_t
	{
		virtual bool render(wxDC& oDest, wxDC* pObjectSheet);
		virtual void load(const char* sData, size_t iLen);

		int x,y, sx, sy, o, o1, o2, u1, u2;
	};

	struct _drone_t : _object_t
	{
		virtual int getSheetIndex();
		virtual bool render(wxDC& oDest, wxDC* pObjectSheet);
		virtual void load(const char* sData, size_t iLen);

		int x,y,algo,seeker,type,dir;
	};

	std::vector<_object_t*> m_vObjects;
};

#endif