/* 
	Corsix's N Level Generator
    Copyright (C) 2007 Peter Cawley

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <wx/clipbrd.h>
#include "frmMain.h"
#include "LuaGenerator.h"

BEGIN_EVENT_TABLE(frmMain, wxFrame)
	EVT_BUTTON(IDC_Generate, OnGenerate)
	EVT_BUTTON(IDC_GenerateSlowly, OnGenerateSlow)
	EVT_BUTTON(IDC_CopyData, OnCopyData)
	EVT_BUTTON(IDC_AddUserlevel, OnAddUserlevel)
END_EVENT_TABLE()

BEGIN_EVENT_TABLE(BitmapControl, wxPanel)
	EVT_PAINT(OnPaint)
	EVT_ERASE_BACKGROUND(OnErase)
END_EVENT_TABLE()

BitmapControl::BitmapControl(wxWindow* pParent, wxBitmap& bmpBitmap)
	: m_bmpBitmap(bmpBitmap)
{
	wxSize oSize(bmpBitmap.GetWidth(), bmpBitmap.GetHeight());
	Create(pParent, wxID_ANY, wxDefaultPosition, oSize);
}

void BitmapControl::OnPaint(wxPaintEvent& event)
{
	wxPaintDC dcOutput(this);
	wxMemoryDC temp_dc;
	temp_dc.SelectObject(m_bmpBitmap);

	wxRegionIterator upd(GetUpdateRegion());
	while (upd)
	{
		dcOutput.Blit(upd.GetX(), upd.GetY(), upd.GetW(), upd.GetH(), &temp_dc, upd.GetX(), upd.GetY());
		++upd;
	}
}

void BitmapControl::OnErase(wxEraseEvent& event)
{
}

frmMain::frmMain(const wxString& title, const wxPoint& pos, const wxSize& size)
	: wxFrame((wxFrame *)NULL, wxID_ANY, title, pos, size, wxDEFAULT_FRAME_STYLE & ~wxRESIZE_BORDER & ~wxMAXIMIZE_BOX)
	, m_bmpTerrain(Terrain::k_iWidth * Terrain::k_iCellSize, Terrain::k_iHeight * Terrain::k_iCellSize)
{
	wxBoxSizer *pTopSizer = new wxBoxSizer( wxVERTICAL );

	wxStaticBoxSizer* pPreviewSizer = new wxStaticBoxSizer( wxHORIZONTAL, this, wxT("Map Preview") );
	pPreviewSizer->Add(m_pPreviewWindow = new BitmapControl(this, m_bmpTerrain), 0, wxALL | wxFIXED_MINSIZE, 2);
	pTopSizer->Add(pPreviewSizer, 0, wxEXPAND | wxALIGN_LEFT);

	wxStaticBoxSizer *pDataContainer = new wxStaticBoxSizer( wxVERTICAL, this, wxT("Attributes") );
	wxFlexGridSizer* pDataGrid = new wxFlexGridSizer(2);
	pDataGrid->SetFlexibleDirection(wxHORIZONTAL);
	pDataGrid->AddGrowableCol(1, 1);

	pDataGrid->Add(new wxStaticText(this, wxID_ANY, wxT("Name")), 0, wxALIGN_CENTER_VERTICAL | wxFIXED_MINSIZE | wxALL, 2);
	pDataGrid->Add(m_txtMapName = new wxTextCtrl(this, wxID_ANY, wxT("empty map")), 1, wxEXPAND | wxALIGN_CENTER_VERTICAL | wxALL, 2);

	pDataGrid->Add(new wxStaticText(this, wxID_ANY, wxT("Author")), 0, wxALIGN_CENTER_VERTICAL | wxFIXED_MINSIZE | wxALL, 2);
	pDataGrid->Add(m_txtMapAuthor = new wxTextCtrl(this, wxID_ANY, wxT("nobody")), 1, wxEXPAND | wxALIGN_CENTER_VERTICAL | wxALL, 2);

	pDataGrid->Add(new wxStaticText(this, wxID_ANY, wxT("Data")), 0, wxALIGN_CENTER_VERTICAL | wxFIXED_MINSIZE | wxALL, 2);
	pDataGrid->Add(m_txtMapData = new wxTextCtrl(this, wxID_ANY, wxString((wxChar)'0', (Terrain::k_iWidth - 2) * (Terrain::k_iHeight - 2))), 1, wxEXPAND | wxALIGN_CENTER_VERTICAL | wxALL, 2);

	pDataContainer->Add(pDataGrid, 0, wxEXPAND | wxALIGN_LEFT);
	pTopSizer->Add(pDataContainer, 0, wxEXPAND | wxALIGN_LEFT);

	wxStaticBoxSizer *pButtonSizer = new wxStaticBoxSizer( wxHORIZONTAL, this, wxT("Commands") );
	pButtonSizer->Add(m_btnGenerate = new wxButton(this, IDC_Generate, wxT("Generate New Level")), 0, wxALL, 2);
	pButtonSizer->Add(m_btnGenerateSlow = new wxButton(this, IDC_GenerateSlowly, wxT("Generate Slowly")), 0, wxALL, 2);
	pButtonSizer->Add(new wxButton(this, IDC_CopyData, wxT("Copy Data to Clipboard")), 0, wxALL, 2);
	pButtonSizer->Add(new wxButton(this, IDC_AddUserlevel, wxT("Add to Userlevels")), 0, wxALL, 2);
	pTopSizer->Add(pButtonSizer, 0, wxEXPAND | wxALIGN_LEFT);

	SetSizer(pTopSizer);
	pTopSizer->SetSizeHints( this );

	m_oTerrain.generateTileImages();
	m_oTerrain.loadObjectImages();
	_renderTerrainToBitmap();

	SetBackgroundColour(m_btnGenerate->GetBackgroundColour());
}

void frmMain::OnCopyData(wxCommandEvent& e)
{
	if(wxTheClipboard->Open())
	{
		wxTheClipboard->SetData(new wxTextDataObject(m_txtMapData->GetValue()));
		wxTheClipboard->Close();
	}
}

void frmMain::OnAddUserlevel(wxCommandEvent& e)
{
	FILE *f = fopen("userlevels.txt", "r+");
	if(f == 0)
	{
		wxString sUserlevels = ::wxFileSelector(wxT("Locate Userlevels.txt"), wxT(""), wxT("userlevels.txt"), wxT("txt"), wxT("Text files (*.txt)|*.txt|All files (*.*)|*.*"), wxOPEN | wxFILE_MUST_EXIST, this);
		if(sUserlevels.IsEmpty())
			return;
		f = fopen(sUserlevels.mb_str(), "r+");
		if(f == 0)
		{
			::wxMessageBox(wxT("Unable to load userlevels.txt"), wxT("Add to Userlevels"), wxICON_ERROR, this);
			return;
		}
	}
	wxString sData;
	sData += wxT("\n$");
	sData += m_txtMapName->GetValue();
	sData += wxT("#");
	sData += m_txtMapAuthor->GetValue();
	sData += wxT("#Action#");
	sData += m_txtMapData->GetValue();
	sData += wxT("#\n");

	fseek(f, 0, SEEK_END);
	fwrite(sData.mb_str(), 1, strlen(sData.mb_str()), f);
	fclose(f);
}

bool frmMain::_renderTerrainToBitmap()
{
	wxMemoryDC oBitmapDC;
	wxColour clrBackground(202, 202, 208);
	wxBrush oBackgroundBrush(clrBackground);

	oBitmapDC.SelectObject(m_bmpTerrain);

	oBitmapDC.SetBackground(oBackgroundBrush);
	oBitmapDC.Clear();

	return m_oTerrain.renderOnto(oBitmapDC);
}

void frmMain::setData(const char* sNewData)
{
	m_oTerrain.load(sNewData);
	_renderTerrainToBitmap();

	m_pPreviewWindow->Refresh();
	m_txtMapData->ChangeValue( wxString(sNewData, wxConvUTF8) );

	wxSafeYield();
}

void frmMain::OnGenerateSlow(wxCommandEvent& e)
{
	m_btnGenerate->Enable(false);
	m_btnGenerateSlow->Enable(false);

	LuaGenerator oGen;
	if(!oGen.init("lua/generator.lua", this))
	{
		::wxMessageBox(wxString(wxT("Error initialising the generator: ")) + oGen.getErrorDetail(), wxT("Generate New Level"), wxICON_ERROR | wxOK, this);
	}
	else
	{
		oGen.setSpeed(1);
		if(!oGen.generate(this))
		{
			::wxMessageBox(wxString(wxT("Error running the generator: ")) + oGen.getErrorDetail(), wxT("Generate New Level"), wxICON_ERROR | wxOK, this);
		}
	}

	m_btnGenerate->Enable(true);
	m_btnGenerateSlow->Enable(true);
}

void frmMain::OnGenerate(wxCommandEvent& e)
{
	LuaGenerator oGen;
	if(!oGen.init("lua/generator.lua", this))
	{
		::wxMessageBox(wxString(wxT("Error initialising the generator: ")) + oGen.getErrorDetail(), wxT("Generate New Level"), wxICON_ERROR | wxOK, this);
		return;
	}
	if(!oGen.generate(this))
	{
		::wxMessageBox(wxString(wxT("Error running the generator: ")) + oGen.getErrorDetail(), wxT("Generate New Level"), wxICON_ERROR | wxOK, this);
		return;
	}
}